Unit LockRight;

Interface

Uses
  Tomas,
  SysUtils;


Type

  TLockKeeper = class;
  TShip = class;
  TGenerator = class;
  TLock = class;

  SideType=(Left,Right);

  TShip = class(TomasElement)
    public
    ShipLength: Double;
    ShipSide: SideType;
    Published
    Constructor Create(Nm: String; L: Double; Side: SideType);
    Procedure Process; override;
    end;

  TLockKeeper = class(TomasElement)
    Published
    Procedure Process; override;
    end;

  TGenerator = class(TomasElement)
    public
    InterArrivalTimes: TExponentialDistribution;
    ShipLengths: TUniformDistribution;
    Published
    Constructor Create(Nm: String);
    Procedure Process; override;
    End;

  TLock = class
    public
    Q: Array[Left..Right] Of TomasQueue;
    QLock: TomasQueue;
    LockSide: SideType;
    Constructor Create;
    End;

Var
  Generator: TGenerator;
  Lock: TLock;
  LockKeeper: TLockKeeper;

implementation

Procedure TLockKeeper.Process;
  Var
    FirstShip: TShip;
  Begin
  While TRUE Do
    Begin
    While (Lock.Q[Left].Length + Lock.Q[Right].Length = 0) Do
      Begin
      Standby;
      End;
    If Lock.Q[Lock.LockSide].Length = 0 Then
      Begin
      Hold(10);
      Lock.LockSide :=  SideType(1 - Ord(Lock.LockSide));
      End;
    While Lock.Q[Lock.LockSide].Length > 0 Do
      Begin
      FirstShip:=Lock.Q[Lock.LockSide].FirstElement;
      FirstShip.Resume(TNow);
      Suspend;
      End;
    Hold(10);
    Lock.LockSide :=  SideType(1 - Ord(Lock.LockSide));
    While Lock.QLock.Length > 0 Do
      Begin
      FirstShip:=Lock.QLock.FirstElement;
      FirstShip.Resume(TNow);
      Suspend;
      End;
    End;
  End;

Constructor TShip.Create(Nm: String; L:Double; Side: SideType);
  Begin
  inherited Create(Nm);
  ShipLength:=L;
  ShipSide:=Side;
  End;

Procedure TShip.Process;
  Begin
  EnterQueue(Lock.Q[ShipSide]);
  Suspend;
  LeaveQueue(Lock.Q[ShipSide]);
  Hold(2);
  EnterQueue(Lock.QLock);
  LockKeeper.Resume(TNow);
  Suspend;
  LeaveQueue(Lock.QLock);
  Hold(2);
  LockKeeper.Resume(TNow);
  If ShipSide = Right Then
    Begin
    SendMessageToModel('plockleft',Name + '!!' + FormatFloat('0.00',ShipLength)
     + '!!' + FormatFloat('0.00',TNow + 20));
    End;
  StopAndLeave;
  End;

Constructor TGenerator.Create(Nm: String);
  Begin
  inherited Create(Nm);
  InterArrivalTimes:=TExponentialDistribution.Create(520623,40);
  ShipLengths:=TUniformDistribution.Create(31659442,15,30);
  End;

Procedure TGenerator.Process;
  Var
    NewShip: TShip;
  Begin
  While TRUE Do
    Begin
    NewShip:=TShip.Create('ShipR',ShipLengths.Sample,Right);
    NewShip.Start(TNow);
    Hold(InterArrivalTimes.Sample);
    End;
  End;

Constructor TLock.Create;
  Begin
  Q[Left]:=TomasQueue.Create('LeftArrivals');
  Q[Right]:=TomasQueue.Create('RightArrivals');
  QLock:=TomasQueue.Create('Lock');
  LockSide:=Left;
  End;

Procedure ReceiveShipsFromLeft(Msg: String);
  Var
    Q: Integer;
    Nm: String;
    L: Double;
    T: Double;
    NewShip: TShip;
    MsgL: TMsgString;
  Begin
  If Pos('$$from:',Msg) <> 0 Then   {must be a ship from the right}
    Begin
    Q:=Pos('!!',Msg);  {after first !! the real message starts}
    Msg:=Trim(Copy(Msg,Q+2,Length(Msg)-Q-1));
    MsgL:=TmsgString.Create(Msg);
    {msgL is now 'shipid!!shiplength!!arrivaltime-at-lock'}
    Nm:=MsgL.GetString;
    L:=MsgL.GetDouble;
    T:=MsgL.GetDouble;
    MsgL.Free;
    NewShip:=TShip.Create(Nm,L,Left);
    NewShip.Start(T);
    End;
  End;


Procedure Acknowledge(Q: TMsgString; Source: String);
  begin
  SendReply(Source,'Yes');
  End;

Procedure InitializeLock;
  Begin
  Lock:=TLock.Create;
  Generator:=TGenerator.Create('Generator');
  Generator.Start(TNow);
  LockKeeper:=TLockKeeper.Create('LockKeeper');
  LockKeeper.Start(TNow);
  TomasForm.SetTrace(TRUE);
  (* the following 3 statements guarantee a correct connection to the server *)
  SetServerName('LocalHost');
  SetReceive(ReceiveShipsFromLeft);
  SetAnswerProc(Acknowledge);
  (***************************************************************************)
  End;


Initialization
  InitializeLock;

end.
